#include "G65DDCAPI.h"
#include <stdio.h>
#include <mmsystem.h>
#include <math.h>

#define MSG_EXITTHREAD (WM_APP+1000)

G65DDC_OPEN_DEVICE OpenDevice;
G65DDC_CLOSE_DEVICE CloseDevice;
G65DDC_SET_POWER SetPower;
G65DDC_CREATE_DDC1 CreateDDC1;
G65DDC_CREATE_DDC2 CreateDDC2;
G65DDC_GET_RANGE GetRange;
G65DDC_SET_DDC SetDDC;
G65DDC_GET_DDC GetDDC;
G65DDC_SET_CALLBACKS SetCallbacks;
G65DDC_SET_FREQUENCY SetFrequency;
G65DDC_GET_FREQUENCY GetFrequency;
G65DDC_START_DDC StartDDC;
G65DDC_STOP_DDC StopDDC;
G65DDC_START_AUDIO StartAudio;
G65DDC_STOP_AUDIO StopAudio;
G65DDC_SET_AGC SetAGC;
G65DDC_SET_AGC_PARAMS SetAGCParams;
G65DDC_SET_DEMODULATOR_MODE SetDemodulatorMode;
G65DDC_SET_DEMODULATOR_FILTER_BANDWIDTH SetDemodulatorFilterBandwidth;
G65DDC_SET_DDC_FREQUENCY SetDDCFrequency;
G65DDC_SET_DEMODULATOR_FREQUENCY SetDemodulatorFrequency;
G65DDC_GET_DDC_FREQUENCY GetDDCFrequency;
G65DDC_GET_DEMODULATOR_FREQUENCY GetDemodulatorFrequency;


INT32 hDevice;

HMODULE hAPI=NULL;
HANDLE hThread;
DWORD ThreadId;
HWAVEOUT hWaveOut;
UINT32 BufferCount;


//procedure of thread used for freeing audio buffers returned to application
//from audio subsystem
DWORD __stdcall WaveOutThreadProcedure(void *Param)
{
 MSG Msg;
 WAVEHDR *WaveHdr;

    while(GetMessage(&Msg,NULL,0,0))
    {
        switch(Msg.message)
        {
            case MSG_EXITTHREAD:
                //exit thread
                return 0;

            case MM_WOM_DONE:
                WaveHdr=(WAVEHDR*)Msg.lParam;

                waveOutUnprepareHeader(hWaveOut,WaveHdr,sizeof(WAVEHDR));

                VirtualFree(WaveHdr,0,MEM_RELEASE);

                BufferCount--;
                break;

            default:
                DispatchMessage(&Msg);
                break;
        }
    }

    return 0;
}

BOOL Initialize(void)
{
 WAVEFORMATEX WaveFormat;

    hAPI=LoadLibrary("G65DDCAPI.dll");

    if(hAPI==NULL)
    {
        printf("Failed to load G65DDCAPI.dll.\n");
        return FALSE;
    }

    //retrieve addresses of used API functions
    OpenDevice=(G65DDC_OPEN_DEVICE)GetProcAddress(hAPI,"OpenDevice");
    CloseDevice=(G65DDC_CLOSE_DEVICE)GetProcAddress(hAPI,"CloseDevice");
    SetPower=(G65DDC_SET_POWER)GetProcAddress(hAPI,"SetPower");
    CreateDDC1=(G65DDC_CREATE_DDC1)GetProcAddress(hAPI,"CreateDDC1");
    CreateDDC2=(G65DDC_CREATE_DDC2)GetProcAddress(hAPI,"CreateDDC2");
    GetRange=(G65DDC_GET_RANGE)GetProcAddress(hAPI,"GetRange");
    SetDDC=(G65DDC_SET_DDC)GetProcAddress(hAPI,"SetDDC");
    GetDDC=(G65DDC_GET_DDC)GetProcAddress(hAPI,"GetDDC");
    SetCallbacks=(G65DDC_SET_CALLBACKS)GetProcAddress(hAPI,"SetCallbacks");
    SetFrequency=(G65DDC_SET_FREQUENCY)GetProcAddress(hAPI,"SetFrequency");
    GetFrequency=(G65DDC_GET_FREQUENCY)GetProcAddress(hAPI,"GetFrequency");
    StartDDC=(G65DDC_START_DDC)GetProcAddress(hAPI,"StartDDC");
    StopDDC=(G65DDC_STOP_DDC)GetProcAddress(hAPI,"StopDDC");
    StartAudio=(G65DDC_START_AUDIO)GetProcAddress(hAPI,"StartAudio");
    StopAudio=(G65DDC_STOP_AUDIO)GetProcAddress(hAPI,"StopAudio");
    SetAGC=(G65DDC_SET_AGC)GetProcAddress(hAPI,"SetAGC");
    SetAGCParams=(G65DDC_SET_AGC_PARAMS)GetProcAddress(hAPI,"SetAGCParams");
    SetDemodulatorMode=(G65DDC_SET_DEMODULATOR_MODE)GetProcAddress(hAPI,"SetDemodulatorMode");
    SetDemodulatorFilterBandwidth=(G65DDC_SET_DEMODULATOR_FILTER_BANDWIDTH)GetProcAddress(hAPI,"SetDemodulatorFilterBandwidth");
    SetDDCFrequency=(G65DDC_SET_DDC_FREQUENCY)GetProcAddress(hAPI,"SetDDCFrequency");
    SetDemodulatorFrequency=(G65DDC_SET_DEMODULATOR_FREQUENCY)GetProcAddress(hAPI,"SetDemodulatorFrequency");
    GetDDCFrequency=(G65DDC_GET_DDC_FREQUENCY)GetProcAddress(hAPI,"GetDDCFrequency");
    GetDemodulatorFrequency=(G65DDC_GET_DEMODULATOR_FREQUENCY)GetProcAddress(hAPI,"GetDemodulatorFrequency");

    //open the first available G65DDC device
    hDevice=OpenDevice(G65DDC_OPEN_FIRST);

    if(hDevice<0)
    {
        printf("No available G65DDC device found.\n");
        return FALSE;
    }


    //create thread which frees audio buffers returned to application
    hThread=CreateThread(NULL,0,WaveOutThreadProcedure,NULL,0,&ThreadId);

    if(hThread==NULL)
    {
        printf("Failed to create thread. Error=%u\n",GetLastError());
        CloseDevice(hDevice);
        FreeLibrary(hAPI);
        return FALSE;
    }

    //initialize waveout
    WaveFormat.wFormatTag=WAVE_FORMAT_PCM;
    WaveFormat.nChannels=2;
    WaveFormat.nSamplesPerSec=48000;
    WaveFormat.wBitsPerSample=16;
    WaveFormat.nBlockAlign=WaveFormat.nChannels*WaveFormat.wBitsPerSample/8;
    WaveFormat.nAvgBytesPerSec=WaveFormat.nBlockAlign*WaveFormat.nSamplesPerSec;
    WaveFormat.cbSize=0;

    //reset buffer count
    BufferCount=0;

    if(waveOutOpen(&hWaveOut,WAVE_MAPPER,&WaveFormat,ThreadId,NULL,CALLBACK_THREAD)!=MMSYSERR_NOERROR)
    {
        printf("Failed to open waveout\n");
        PostThreadMessage(ThreadId,MSG_EXITTHREAD,0,0);
        WaitForSingleObject(hThread,INFINITE);
        CloseHandle(hThread);
        CloseDevice(hDevice);
        FreeLibrary(hAPI);
        return FALSE;
    }

    return TRUE;
}

void __stdcall MyDDC2PreprocessedStreamCallback(UINT32 DDC2ChannelId,const float *Buffer,UINT32 Count,float SlevelPeak,float SlevelRMS,DWORD_PTR UserData)
{
 double Slevel_dBm;

    //convert slevel RMS in Volts to dBm
    Slevel_dBm=10.0*log10(SlevelRMS*SlevelRMS*(1000.0/50.0));

    printf("\rSlevel: RMS[V]=%9.7f V, RMS[dBm]=%6.1f dBm, Peak[V]=%9.7f V",SlevelRMS,Slevel_dBm,SlevelPeak);
}

void __stdcall MyAudioStreamCallback(UINT32 DDC2ChannelId,UINT32 Stage,const float *Buffer,UINT32 Count,DWORD_PTR UserData)
{
 WAVEHDR *WaveHdr;
 SHORT *Output;

    if(Stage==G65DDC_AUDIO_STREAM_CALLBACK_STAGE_2)
    {
        //alocate memory for waveout audio buffer and WAVEHDR structure
        WaveHdr=(WAVEHDR*)VirtualAlloc(NULL,sizeof(WAVEHDR)+Count*2*sizeof(SHORT),MEM_COMMIT,PAGE_READWRITE);

        ZeroMemory(WaveHdr,sizeof(WAVEHDR));

        WaveHdr->lpData=(LPSTR)(WaveHdr+1); //audio buffer starts after WAVEHDR structure
        WaveHdr->dwBufferLength=Count*2*sizeof(SHORT);

        //retype audio buffer ptr to signed 16bit
        Output=(SHORT*)WaveHdr->lpData;

        //convert samples from 32-bit IEEE float to 16-bit integer
        Count*=2;//signal consists of two channels
        while(Count--)
        {
            *Output++=(SHORT)((*Buffer++)*32767.0);
        }

        waveOutPrepareHeader(hWaveOut,WaveHdr,sizeof(WAVEHDR));

        //increase number of buffers sent to waveout
        BufferCount++;

        if(waveOutWrite(hWaveOut,WaveHdr,sizeof(WAVEHDR))!=MMSYSERR_NOERROR)
        {
            //decrease number of buffers sent to waveout because an error
            BufferCount--;
            VirtualFree(WaveHdr,0,MEM_RELEASE);
        }
    }
}

int main(int argc, char* argv[])
{
 G65DDC_DDC_INFO DDCInfo;
 UINT32 DDCTypeIndex;
 G65DDC_CALLBACKS Callbacks;
 INT32 Frequency;
 INT32 DDC1Frequency;
 INT32 DDC2Frequency;
 INT32 DemodulatorFrequency;
 UINT32 Range;
 UINT32 DDC1Id;
 UINT32 DDC2Id;

    if(!Initialize())
    {
        printf("\nPress ENTER to exit.\n");
        getchar();
        return -1;
    }

    //create DDC1
    CreateDDC1(hDevice,&DDC1Id);

    //create DDC2
    CreateDDC2(hDevice,DDC1Id,&DDC2Id);

    //set power
    SetPower(hDevice,TRUE);

    //set DDC type of DDC1 to 1 (DDC Bandwidth=24000, SampleRate=32kHz, BitsPerSample=32)
    SetDDC(hDevice,DDC1Id,1);

    //retrieve current DDC type of the DDC1
    GetDDC(hDevice,DDC1Id,&DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC1: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);

    //retrieve current DDC type of the DDC2
    GetDDC(hDevice,DDC2Id,&DDCTypeIndex,&DDCInfo);
    printf("DDC type of DDC2: Index=%u, Bandwidth=%u, SampleRate=%u, BitsPerSample=%u\n",DDCTypeIndex,DDCInfo.Bandwidth,DDCInfo.SampleRate,DDCInfo.BitsPerSample);

    //ADCSnapshotCallback is not required in this example
    Callbacks.ADCSnapshotCallback=NULL;

    //DDC1StreamCallback is not required in this example
    Callbacks.DDC1StreamCallback=NULL;

    //DDC1PlaybackStreamCallback is not required in this example
    Callbacks.DDC1PlaybackStreamCallback=NULL;

    //DDC2StreamCallback is not required in this example
    Callbacks.DDC2StreamCallback=NULL;

    //DDC2StreamCallback is not required in this example
    Callbacks.DDC2StreamCallback=NULL;

    //DDC2PreprocessedStreamCallback is used to display signal level in this example
    Callbacks.DDC2PreprocessedStreamCallback=MyDDC2PreprocessedStreamCallback;

    //AudioStreamCallback is used to play audio on wave out device
    Callbacks.AudioStreamCallback=MyAudioStreamCallback;

    //AudioPlaybackStreamCallback is not used in this example
    Callbacks.AudioPlaybackStreamCallback=NULL;

    //register callback functiobs
    SetCallbacks(hDevice,&Callbacks,0);

    //tune to an AM station frequency
    SetFrequency(hDevice,DDC2Id,702000);
    //it is possible to tune DDC1, DDC2 and demodulator directly using
    //SetDDCFrequency and SetDemodulatorFrequency
    //SetFrequency(hDevice,DDC2Id,702000) can be replaced by the following:
    //SetRange(hDevice,G65DDC_RANGE_1);
    //SetDDCFrequency(hDevice,DDC1Id,700000);
    //SetDDCFrequency(hDevice,DDC2Id,0);
    //SetDemodulatorFrequency(hDevice,DDC2Id,2000);
    //final absolute frequency of the demodulator is 700000+0+2000=702000

    //get absolute frequency of the demodulator
    GetFrequency(hDevice,DDC2Id,&Frequency);

    //get range
    GetRange(hDevice,&Range);

    //get DDC1 frequency
    GetDDCFrequency(hDevice,DDC1Id,&DDC1Frequency);

    //get relative DDC2 frequency
    GetDDCFrequency(hDevice,DDC2Id,&DDC2Frequency);

    //get relative demodulator frequency
    GetDemodulatorFrequency(hDevice,DDC2Id,&DemodulatorFrequency);

    printf("Listening to (demodulator absolute frequency): %u Hz\n",Frequency);
    printf("Range: %u\n",Range);
    printf("DDC1 frequency: %d Hz\n",DDC1Frequency);
    printf("DDC2 frequency (relative): %d Hz\n",DDC2Frequency);
    printf("Demodulator frequency (relative): %d Hz\n",DemodulatorFrequency);


    //set AGC params, attack=0.02s, decay=0.5s, ref. level=-15dB
    SetAGCParams(hDevice,DDC2Id,0.020,0.500,-15.0);

    //enable AGC for channel 0
    SetAGC(hDevice,0,TRUE);

    //set AM mode
    SetDemodulatorMode(hDevice,DDC2Id,G65DDC_MODE_AM);

    //set demodulator filter bandwidth to 8kHz
    SetDemodulatorFilterBandwidth(hDevice,DDC2Id,8000);

    //start DDC1 streaming
    StartDDC(hDevice,DDC1Id,1024);

    //start DDC2 streaming
    StartDDC(hDevice,DDC2Id,1024);

    //start audio streaming, 1024 samples per buffer
    StartAudio(hDevice,DDC2Id,1024);

    printf("\nPress ENTER to exit.\n\n");
    getchar();

    //stop audio streaming
    StopAudio(hDevice,DDC2Id);

    //stop DDC2 streaming
    StopDDC(hDevice,DDC2Id);

    //stop DDC1 streaming
    StopDDC(hDevice,DDC1Id);

    waveOutReset(hWaveOut);

    //wait for all buffers sent to waveout are returned
    while(BufferCount)
    {
        Sleep(10);
    }

    PostThreadMessage(ThreadId,MSG_EXITTHREAD,0,0);
    WaitForSingleObject(hThread,INFINITE);
    CloseHandle(hThread);

    waveOutClose(hWaveOut);

    //unregister all callback functions
    SetCallbacks(hDevice,NULL,0);

    SetPower(hDevice,FALSE);

    CloseDevice(hDevice);

    FreeLibrary(hAPI);

    return 0;
}

